# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import pathlib

import pytest
from capellambse.filehandler import abc

from capella_ros_tools import data_model
from capella_ros_tools.data_model import (
    CONSTANT_SEPARATOR,
    UPPER_BOUND_TOKEN,
    ConstantDef,
    EnumDef,
    FieldDef,
    MessageDef,
    MessagePkgDef,
    Range,
    TypeDef,
)

PATH = pathlib.Path(__file__).parent

SAMPLE_CLASS_PATH = PATH.joinpath(
    "data/data_model/example_msgs/package1/msg/SampleClass.msg"
)
SAMPLE_ENUM_PATH = PATH.joinpath(
    "data/data_model/example_msgs/package1/msg/types/SampleEnum.msg"
)
SAMPLE_CLASS_ENUM_PATH = PATH.joinpath(
    "data/data_model/example_msgs/package2/msg/SampleClassEnum.msg"
)

SAMPLE_PACKAGE_PATH1 = PATH.joinpath("data/data_model/example_msgs/package1")
SAMPLE_PACKAGE_PATH2 = PATH.joinpath("data/data_model/example_msgs/package2")


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            ("test_name", Range("1", "1"), "test_package"),
            "test_package/test_name",
        ),
        (
            ("test_name", Range("10", "10"), "test_package"),
            "test_package/test_name[10]",
        ),
        (
            ("test_name", Range("0", "10"), "test_package"),
            f"test_package/test_name[{UPPER_BOUND_TOKEN}10]",
        ),
        (
            ("test_name", Range("1", "1"), None),
            "test_name",
        ),
        (
            ("test_name", Range("0", "*"), None),
            "test_name[]",
        ),
    ],
)
def test_TypeDef_str(params: tuple[str, Range, str | None], expected: str):
    type_def = TypeDef(*params)

    actual = str(type_def)

    assert actual == expected


@pytest.mark.parametrize(
    "type_str, params",
    [
        (
            "test_package/test_name",
            ("test_name", Range("1", "1"), "test_package"),
        ),
        (
            "test_package/test_name[10]",
            ("test_name", Range("10", "10"), "test_package"),
        ),
        (
            f"test_package/test_name[{UPPER_BOUND_TOKEN}10]",
            ("test_name", Range("0", "10"), "test_package"),
        ),
        (
            "test_name[]",
            ("test_name", Range("0", "*"), None),
        ),
    ],
)
def test_TypeDef_from_string(
    type_str: str, params: tuple[str, Range, str | None]
):
    expected = TypeDef(*params)

    actual = TypeDef.from_string(type_str)

    assert actual == expected


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            (
                TypeDef.from_string("test_type"),
                "test_name",
                "test_description",
            ),
            "test_type test_name    # test_description",
        ),
        (
            (TypeDef.from_string("test_type"), "test_name", ""),
            "test_type test_name",
        ),
    ],
)
def test_FieldDef_str(params: tuple[TypeDef, str, str], expected: str):
    field_def = FieldDef(*params)

    actual = str(field_def)

    assert actual == expected


@pytest.mark.parametrize(
    "params, expected",
    [
        (
            (
                TypeDef.from_string("test_type"),
                "test_name",
                "1",
                "test",
            ),
            f"test_type test_name {CONSTANT_SEPARATOR} 1    # test",
        ),
        (
            (TypeDef.from_string("test_type"), "test_name", "10", ""),
            f"test_type test_name {CONSTANT_SEPARATOR} 10",
        ),
    ],
)
def test_ConstantDef_str(params: tuple[TypeDef, str, str, str], expected: str):
    constant_def = ConstantDef(*params)

    actual = str(constant_def)

    assert actual == expected


class TestComments:
    @staticmethod
    def test_extract_file_level_comments_no_comments():
        msg_string = """uint8 OK = 0
uint8 WARN = 1
uint8 ERROR = 2"""
        comments, _ = data_model._extract_file_level_comments(msg_string)

        assert comments == ""

    @staticmethod
    def test_extract_file_level_comments_no_newline():
        msg_string = """# This is a comment
# This is another comment
uint8 OK = 0
uint8 WARN = 1
uint8 ERROR = 2"""
        comments, _ = data_model._extract_file_level_comments(msg_string)

        assert comments == ""

    @staticmethod
    def test_extract_file_level_comments():
        msg_string = """# This is a comment
# This is another comment

uint8 OK = 0
uint8 WARN = 1
uint8 ERROR = 2"""
        comments, _ = data_model._extract_file_level_comments(msg_string)
        expected = "This is a comment This is another comment "

        assert comments == expected

    @staticmethod
    def test_extract_file_level_comments_with_newline():
        msg_string = """# This is a comment
#
# This is another comment"""
        comments, _ = data_model._extract_file_level_comments(msg_string)
        expected = "This is a comment <br>This is another comment "

        assert comments == expected

    @staticmethod
    def test_extract_file_level_comments_strip_empty_lines_at_top():
        msg_string = """

# This is a comment

uint8 OK = 0
uint8 WARN = 1
uint8 ERROR = 2"""
        comments, _ = data_model._extract_file_level_comments(msg_string)
        expected = "This is a comment "

        assert comments == expected

    @staticmethod
    def test_parse_comments_no_comments():
        msg_string = """uint8 field"""
        msg_def = MessageDef.from_string("test_name", msg_string)
        expected = MessageDef(
            name="test_name",
            fields=[
                FieldDef(
                    type=TypeDef.from_string("uint8"),
                    name="field",
                    description="",
                )
            ],
            enums=[],
            description="",
        )

        assert msg_def == expected

    @staticmethod
    def test_parse_comments_block_comments():
        msg_string = """# Here is text.
# Here is more text.
#
# This is unrelated text.
uint8 field"""
        msg_def = MessageDef.from_string("test_name", msg_string)
        expected = MessageDef(
            name="test_name",
            fields=[
                FieldDef(
                    type=TypeDef.from_string("uint8"),
                    name="field",
                    description="Here is text. Here is more text. <br>"
                    "This is unrelated text. ",
                )
            ],
            enums=[],
            description="",
        )

        assert msg_def == expected

    @staticmethod
    def test_parse_comments_inline_comments():
        msg_string = """uint8 field     # Here is text.
                                        # Here is more text.
                                        #
                                        # This is unrelated text."""
        msg_def = MessageDef.from_string("test_name", msg_string)
        expected = MessageDef(
            name="test_name",
            fields=[
                FieldDef(
                    type=TypeDef.from_string("uint8"),
                    name="field",
                    description="Here is text. Here is more text. <br>"
                    "This is unrelated text. ",
                )
            ],
            enums=[],
            description="",
        )

        assert msg_def == expected

    @staticmethod
    def test_parse_comments_mixed_comments():
        msg_string = """# This is a block comment.
# This is still a block comment.
uint8 field     # This is an inline comment.
                # This is still an inline comment."""
        msg_def = MessageDef.from_string("test_name", msg_string)
        expected = MessageDef(
            name="test_name",
            fields=[
                FieldDef(
                    type=TypeDef.from_string("uint8"),
                    name="field",
                    description="This is a block comment. "
                    "This is still a block comment. "
                    "This is an inline comment. "
                    "This is still an inline comment. ",
                )
            ],
            enums=[],
            description="",
        )

        assert msg_def == expected


class TestMergeEnumDef:
    @staticmethod
    @pytest.fixture
    def expected() -> MessageDef:
        type_def = TypeDef("uint8", Range("1", "1"))
        enum_def = EnumDef(
            name="enum_name",
            literals=[
                ConstantDef(
                    type=type_def,
                    name="OK",
                    value="0",
                    description="",
                ),
                ConstantDef(
                    type=type_def,
                    name="WARN",
                    value="1",
                    description="",
                ),
                ConstantDef(
                    type=type_def,
                    name="ERROR",
                    value="2",
                    description="",
                ),
                ConstantDef(
                    type=type_def,
                    name="STALE",
                    value="3",
                    description="",
                ),
            ],
            description="",
        )
        return MessageDef("enum_name", [], [enum_def], "")

    @staticmethod
    def test_merge_enums_before(expected: MessageDef):

        msg_string = """
uint8 OK = 0

uint8 WARN = 1
uint8 ERROR = 2
uint8 STALE = 3"""
        msg_def = MessageDef.from_string("enum_name", msg_string)

        assert msg_def == expected

    @staticmethod
    def test_merge_enums_after(expected: MessageDef):
        msg_string = """
uint8 OK = 0
uint8 WARN = 1
uint8 ERROR = 2

uint8 STALE = 3"""
        msg_def = MessageDef.from_string("enum_name", msg_string)

        assert msg_def == expected

    @staticmethod
    def test_merge_enums_multiple_after(expected: MessageDef):
        msg_string = """
uint8 OK = 0
uint8 WARN = 1

uint8 ERROR = 2

uint8 STALE = 3"""
        msg_def = MessageDef.from_string("enum_name", msg_string)

        assert msg_def == expected

    @staticmethod
    def test_merge_enums_multiple_before(expected: MessageDef):
        msg_string = """
uint8 OK = 0

uint8 WARN = 1

uint8 ERROR = 2
uint8 STALE = 3"""
        msg_def = MessageDef.from_string("enum_name", msg_string)

        assert msg_def == expected


def test_enum_non_ascending_values():
    msg_string = """
uint8 SHAPE_TYPE_UNDEFINED = 0

# Comment block 1

uint8 SHAPE_TYPE_BOX = 1
uint8 SHAPE_TYPE_SPHERE = 2

# Comment block 2

uint8 SHAPE_TYPE_VERTICAL_STRUCTURE = 10
uint8 SHAPE_TYPE_VERTICAL_STRUCTURE_WITH_RADIUS = 101
uint8 SHAPE_TYPE_HORIZONTAL_STRUCTURE = 11"""
    msg_def = MessageDef.from_string("ShapeTypes", msg_string)
    expected = MessageDef(
        name="ShapeTypes",
        fields=[],
        enums=[
            EnumDef(
                name="ShapeTypes",
                literals=[
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1")),
                        name="UNDEFINED",
                        value="0",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1")),
                        name="BOX",
                        value="1",
                        description="Comment block 1 ",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1")),
                        name="SPHERE",
                        value="2",
                        description="Comment block 1 ",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1")),
                        name="VERTICAL_STRUCTURE",
                        value="10",
                        description="Comment block 2 ",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1")),
                        name="VERTICAL_STRUCTURE_WITH_RADIUS",
                        value="101",
                        description="Comment block 2 ",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1")),
                        name="HORIZONTAL_STRUCTURE",
                        value="11",
                        description="Comment block 2 ",
                    ),
                ],
                description="",
            )
        ],
        description="",
    )

    assert msg_def == expected


class TestEnumName:
    @staticmethod
    def test_enum_name_commonprefix_no_underscore():
        msg_string = """
uint8 START = 0
uint8 STOP = 1
int8 field"""
        msg_def = MessageDef.from_string("enum_name", msg_string)
        expected = "enum_nameType"

        actual = msg_def.enums[0].name

        assert actual == expected

    @staticmethod
    def test_enum_name_commonprefix_with_underscore():
        msg_string = """
uint8 ST_ART = 0
uint8 ST_OP = 1
int8 field"""
        msg_def = MessageDef.from_string("enum_name", msg_string)
        expected = "St"

        actual = msg_def.enums[0].name

        assert actual == expected

    @staticmethod
    def test_enum_name_commonprefix_with_multiple_underscore():
        msg_string = """
uint8 S_T_ART = 0
uint8 S_T_OP = 1
int8 field"""
        msg_def = MessageDef.from_string("enum_name", msg_string)
        expected = "ST"

        actual = msg_def.enums[0].name

        assert actual == expected

    @staticmethod
    def test_enum_name_match():
        msg_string = """
int8 STATUS_NO_FIX =  -1
int8 STATUS_FIX =      0
int8 STATUS_SBAS_FIX = 1
int8 STATUS_GBAS_FIX = 2

int8 status

uint16 SERVICE_GPS =     1
uint16 SERVICE_GLONASS = 2
uint16 SERVICE_COMPASS = 4
uint16 SERVICE_GALILEO = 8

uint16 service"""
        msg_def = MessageDef.from_string("NavSatStatus", msg_string)
        expected = MessageDef(
            name="NavSatStatus",
            fields=[
                FieldDef(
                    type=TypeDef("Status", Range("1", "1"), "NavSatStatus"),
                    name="status",
                    description="",
                ),
                FieldDef(
                    type=TypeDef("Service", Range("1", "1"), "NavSatStatus"),
                    name="service",
                    description="",
                ),
            ],
            enums=[
                EnumDef(
                    name="Status",
                    literals=[
                        ConstantDef(
                            type=TypeDef("int8", Range("1", "1")),
                            name="NO_FIX",
                            value="-1",
                            description="",
                        ),
                        ConstantDef(
                            type=TypeDef("int8", Range("1", "1")),
                            name="FIX",
                            value="0",
                            description="",
                        ),
                        ConstantDef(
                            type=TypeDef("int8", Range("1", "1")),
                            name="SBAS_FIX",
                            value="1",
                            description="",
                        ),
                        ConstantDef(
                            type=TypeDef("int8", Range("1", "1")),
                            name="GBAS_FIX",
                            value="2",
                            description="",
                        ),
                    ],
                    description="",
                ),
                EnumDef(
                    name="Service",
                    literals=[
                        ConstantDef(
                            type=TypeDef("uint16", Range("1", "1")),
                            name="GPS",
                            value="1",
                            description="",
                        ),
                        ConstantDef(
                            type=TypeDef("uint16", Range("1", "1")),
                            name="GLONASS",
                            value="2",
                            description="",
                        ),
                        ConstantDef(
                            type=TypeDef("uint16", Range("1", "1")),
                            name="COMPASS",
                            value="4",
                            description="",
                        ),
                        ConstantDef(
                            type=TypeDef("uint16", Range("1", "1")),
                            name="GALILEO",
                            value="8",
                            description="",
                        ),
                    ],
                    description="",
                ),
            ],
            description="",
        )

        assert msg_def == expected


def test_MessageDef_class(sample_class_def: MessageDef):
    msg_path = SAMPLE_CLASS_PATH
    msg_def = MessageDef.from_file(msg_path)
    expected = sample_class_def

    assert msg_def == expected


def test_MessageDef_enum(sample_enum_def: MessageDef):
    msg_path = SAMPLE_ENUM_PATH
    msg_def = MessageDef.from_file(msg_path)
    expected = sample_enum_def

    assert msg_def == expected


def test_MessageDef_class_enum(sample_class_enum_def: MessageDef):
    msg_path = SAMPLE_CLASS_ENUM_PATH
    msg_def = MessageDef.from_file(msg_path)
    expected = sample_class_enum_def

    assert msg_def == expected


@pytest.mark.parametrize(
    "msg_pkg_path",
    [
        SAMPLE_PACKAGE_PATH1,
        SAMPLE_PACKAGE_PATH2,
    ],
)
def test_MessagePkgDef_from_msg_folder(
    msg_pkg_path: abc.AbstractFilePath | pathlib.Path,
):
    message_pkg_def = MessagePkgDef.from_msg_folder("", msg_pkg_path)

    assert message_pkg_def.messages



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Main entry point into Capella ROS Tools."""

import io
import pathlib
import uuid

import capellambse
import click
from capellambse import cli_helpers, decl

import capella_ros_tools
from capella_ros_tools import exporter, importer, logger


@click.group()
@click.version_option(
    version=capella_ros_tools.__version__,
    prog_name="capella-ros-tools",
    message="%(prog)s %(version)s",
)
def cli():
    """Console script for Capella ROS Tools."""


@cli.command("import")
@click.option(
    "-i",
    "--input",
    type=str,
    required=True,
    help="Path to the ROS message package.",
)
@click.option(
    "-m",
    "--model",
    type=cli_helpers.ModelCLI(),
    required=True,
    help="Path to the Capella model.",
)
@click.option(
    "-l",
    "--layer",
    type=click.Choice(["oa", "la", "sa", "pa"], case_sensitive=False),
    help="The layer to import the messages to.",
)
@click.option(
    "-r",
    "--root",
    type=click.UUID,
    help="The UUID of the root package to import the messages to.",
)
@click.option(
    "-t",
    "--types",
    type=click.UUID,
    help="The UUID of the types package to import the created data types to.",
)
@click.option(
    "--no-deps",
    "no_deps",
    is_flag=True,
    help="Don't install message dependencies.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(path_type=pathlib.Path, dir_okay=False),
    help="Produce a declarative YAML instead of modifying the source model.",
)
@click.option(
    "--license-header",
    type=click.Path(path_type=pathlib.Path, dir_okay=False),
    help="Ignore the license header from the given file when importing msgs.",
)
@click.option(
    "--description-regex",
    type=str,
    help="Regular expression to extract description from the file .",
)
def import_msgs(
    input: str,
    model: capellambse.MelodyModel,
    layer: str,
    root: uuid.UUID,
    types: uuid.UUID,
    no_deps: bool,
    output: pathlib.Path,
    license_header: pathlib.Path | None,
    description_regex: str | None,
) -> None:
    """Import ROS messages into a Capella data package."""
    if root:
        root_uuid = str(root)
    elif layer:
        root_uuid = getattr(model, layer).data_package.uuid
    else:
        raise click.UsageError("Either --root or --layer must be provided")

    if types:
        params = {"types_uuid": str(types)}
    else:
        params = {"types_parent_uuid": model.sa.data_package.uuid}

    parsed = importer.Importer(
        input, no_deps, license_header, description_regex
    )
    logger.info("Loaded %d packages", len(parsed.messages.packages))

    yml = parsed.to_yaml(root_uuid, **params)
    if output:
        logger.info("Writing declarative YAML to file %s", output)
        output.write_text(yml, encoding="utf-8")
    else:
        logger.info("Writing to model %s", model.name)
        decl.apply(model, io.StringIO(yml))
        model.save()


@cli.command("export")
@click.option(
    "-m",
    "--model",
    type=cli_helpers.ModelCLI(),
    required=True,
    help="Path to the Capella model.",
)
@click.option(
    "-l",
    "--layer",
    type=click.Choice(["oa", "la", "sa", "pa"], case_sensitive=False),
    help="The layer to export the model objects from.",
)
@click.option(
    "-r",
    "--root",
    type=click.UUID,
    help="The UUID of the root package to import the messages from.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(path_type=pathlib.Path, file_okay=False),
    default=pathlib.Path.cwd() / "data-package",
    help="Output directory for the .msg files.",
)
def export_capella(
    model: capellambse.MelodyModel,
    layer: str,
    root: uuid.UUID,
    output: pathlib.Path,
):
    """Export Capella data package to ROS messages."""
    if root:
        current_pkg = model.search("DataPkg").by_uuid(str(root))
    elif layer:
        current_pkg = getattr(model, layer).data_package
    else:
        raise click.UsageError("Either --root or --layer must be provided")

    exporter.export(current_pkg, output)  # type: ignore


if __name__ == "__main__":
    cli()



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Tool for parsing ROS messages."""

from __future__ import annotations

import os
import pathlib
import re
import typing as t
from dataclasses import dataclass

from capellambse.filehandler import abc

LICENSE_HEADER = (
    pathlib.Path(__file__)
    .parent.joinpath(".license_header.txt")
    .read_text(encoding="utf-8")
)
PACKAGE_NAME_MESSAGE_TYPE_SEPARATOR = "/"
COMMENT_DELIMITER = "#"
CONSTANT_SEPARATOR = "="
UPPER_BOUND_TOKEN = "<="

VALID_MESSAGE_NAME_PATTERN = "[A-Z][A-Za-z0-9]*"
VALID_CONSTANT_NAME_PATTERN = "[A-Z](?:[A-Z0-9_]*[A-Z0-9])?"
VALID_REF_COMMENT_PATTERN = re.compile(
    r".*cf\.\s*"
    rf"({VALID_MESSAGE_NAME_PATTERN})"
    r"(?:,\s*"
    rf"({VALID_CONSTANT_NAME_PATTERN}))?"
    r"\s*.*"
)

HTML_TAG_PATTERN = re.compile("<.*?>")


def _clean_html(raw_html: str):
    return re.sub(HTML_TAG_PATTERN, "", raw_html)


def _clean_comment(comment: str) -> str:
    return comment.strip(COMMENT_DELIMITER).strip()


class Range(t.NamedTuple):
    """Define range of values."""

    min: str
    max: str


@dataclass
class TypeDef:
    """Type definition."""

    name: str
    card: Range
    package: str | None = None

    def __str__(self) -> str:
        """Return string representation of the type."""
        out = self.name
        if self.card.min == self.card.max:
            out += f"[{self.card.max}]" if self.card.max != "1" else ""
        else:
            out += (
                f"[{UPPER_BOUND_TOKEN}{self.card.max}]"
                if self.card.max != "*"
                else "[]"
            )
        if self.package:
            out = f"{self.package}{PACKAGE_NAME_MESSAGE_TYPE_SEPARATOR}{out}"
        return out

    @classmethod
    def from_string(cls, type_str: str) -> TypeDef:
        """Create a type definition from a string."""
        name = type_str
        card = Range("1", "1")
        if type_str.endswith("]"):
            name, _, max_card = type_str.partition("[")
            max_card = max_card.removesuffix("]")
            if max_card.startswith(UPPER_BOUND_TOKEN):
                max_card = max_card.removeprefix(UPPER_BOUND_TOKEN)
                card = Range("0", max_card)
            else:
                card = (
                    Range(max_card, max_card) if max_card else Range("0", "*")
                )

        if len(temp := name.split(PACKAGE_NAME_MESSAGE_TYPE_SEPARATOR)) == 2:
            package, name = temp
        else:
            package = None

        return cls(name, card, package)


@dataclass
class FieldDef:
    """Definition of a field in a ROS message."""

    type: TypeDef
    name: str
    description: str

    def __str__(self) -> str:
        """Return string representation of the field."""
        out = f"{self.type} {self.name}"
        if self.description:
            out += f"    # {_clean_html(self.description)}"
        return out


@dataclass
class ConstantDef:
    """Definition of a constant in a ROS message."""

    type: TypeDef
    name: str
    value: str
    description: str

    def __str__(self) -> str:
        """Return string representation of the constant."""
        out = f"{self.type} {self.name} = {self.value}"
        if self.description:
            out += f"    # {_clean_html(self.description)}"
        return out


@dataclass
class EnumDef:
    """Definition of an enum in a ROS message."""

    name: str
    literals: list[ConstantDef]
    description: str

    def __str__(self) -> str:
        """Return string representation of the enum."""
        out = f"# {_clean_html(self.description)}" if self.description else ""
        for literal in self.literals:
            out += f"\n{literal}"
        return out

    def __eq__(self, other: object) -> bool:
        """Return whether the enum is equal to another."""
        if not isinstance(other, EnumDef):
            return NotImplemented
        return (
            other.name == self.name
            and all(literal in self.literals for literal in other.literals)
            and other.description == self.description
        )


def _process_block_comment(line: str) -> str:
    if comment := _clean_comment(line):
        return f"{comment} "
    return ""


def _extract_file_level_comments(
    msg_string: str, regex: re.Pattern | None = None
) -> t.Tuple[str, list[str]]:
    """Extract comments at the beginning of the message."""
    lines = msg_string.lstrip("\n").splitlines()
    lines.append("")
    file_level_comments = ""
    i = 0
    for i, line in enumerate(lines):
        line = line.strip()
        if not line.startswith(COMMENT_DELIMITER):
            if line:
                return "", lines
            else:
                break
        file_level_comments += _process_block_comment(line) or "\n"

    if regex is not None:
        if matches := regex.search(file_level_comments):
            file_level_comments = "\n".join(matches.groups())
        else:
            file_level_comments = ""
    file_level_comments = file_level_comments.replace("\n", "<br>")
    file_content = lines[i:]
    return file_level_comments, file_content


@dataclass
class MessageDef:
    """Definition of a ROS message."""

    name: str
    fields: list[FieldDef]
    enums: list[EnumDef]
    description: str

    def __str__(self) -> str:
        """Return string representation of the message."""
        if self.description:
            out = f"# {_clean_html(self.description)}\n\n"
        else:
            out = ""
        for enum in self.enums:
            out += f"{enum}\n\n"
        for field in self.fields:
            out += f"{field}\n"
        return out

    def __eq__(self, other: object) -> bool:
        """Return whether the message is equal to another."""
        if not isinstance(other, MessageDef):
            return NotImplemented
        return (
            other.name == self.name
            and all(field in self.fields for field in other.fields)
            and all(enum in self.enums for enum in other.enums)
            and other.description == self.description
        )

    @classmethod
    def from_file(
        cls,
        file: abc.AbstractFilePath | pathlib.Path,
        license_header: str | None = None,
        msg_description_regex: re.Pattern[str] | None = None,
    ) -> MessageDef:
        """Create message definition from a .msg file."""
        msg_name = file.stem
        msg_string = file.read_text()
        license_header = license_header or LICENSE_HEADER
        msg_string = msg_string.removeprefix(license_header)
        return cls.from_string(msg_name, msg_string, msg_description_regex)

    @classmethod
    def from_string(
        cls,
        msg_name: str,
        msg_string: str,
        msg_description_regex: re.Pattern[str] | None = None,
    ) -> MessageDef:
        """Create message definition from a string."""
        msg_comments, lines = _extract_file_level_comments(
            msg_string, msg_description_regex
        )
        msg = cls(msg_name, [], [], msg_comments)
        last_element: t.Any = None
        block_comments = ""
        index = -1
        values: list[str] = []

        for line in lines:
            line = line.rstrip()
            if not line:
                # new block
                if index != 0:
                    block_comments = ""
                continue

            last_index = index
            index = line.find(COMMENT_DELIMITER)
            if index == -1:
                # no comment
                comment = ""
            elif index == 0:
                # block comment
                if last_index > 0:
                    # block comments were used
                    block_comments = ""
                block_comments += _process_block_comment(line) or "<br>"
                continue
            else:
                # inline comment
                comment = _clean_comment(line[index:])
                line = line[:index].rstrip()
                if not line:
                    # indented comment
                    last_element.description += (
                        f"{comment} " if comment else "<br>"
                    )
                    continue
                comment = f"{comment} "

            type_string, _, rest = line.partition(" ")
            name, _, value = rest.partition(CONSTANT_SEPARATOR)
            name = name.strip()
            value = value.strip()
            if value:
                # constant
                if (
                    value in values
                    or not msg.enums
                    or not isinstance(last_element, ConstantDef)
                ):
                    # new enum
                    enum_def = EnumDef("", [], block_comments)
                    block_comments = ""
                    msg.enums.append(enum_def)
                    values = []
                constant_def = ConstantDef(
                    TypeDef.from_string(type_string),
                    name,
                    value,
                    block_comments + comment,
                )
                msg.enums[-1].literals.append(constant_def)
                values.append(value)
                last_element = constant_def
            else:
                # field
                field_def = FieldDef(
                    TypeDef.from_string(type_string),
                    name,
                    block_comments + comment,
                )
                msg.fields.append(field_def)
                last_element = field_def

        if not msg.fields and len(msg.enums) == 1:
            enum = msg.enums[0]
            _process_enums(enum)
            enum.name = msg_name
            return msg

        for field in msg.fields:
            _process_comment(field)

        for enum in msg.enums:

            common_prefix = _process_enums(enum)

            if common_prefix:
                enum.name = _get_enum_identifier(common_prefix)
            else:
                enum.name = msg_name if not msg.fields else msg_name + "Type"

            matched_field = None
            for field in msg.fields:
                if field.type.name == enum.literals[0].type.name:
                    matched_field = matched_field or field
                    if field.name.lower() == enum.name.lower():
                        field.type.name = enum.name
                        field.type.package = msg_name
                        break
            else:
                if matched_field:
                    enum.name = msg_name + matched_field.name.capitalize()
                    matched_field.type.name = enum.name
                    matched_field.type.package = msg_name

        return msg


def _process_enums(enum: EnumDef) -> str:
    common_prefix = os.path.commonprefix(
        [literal.name for literal in enum.literals]
    )
    if not common_prefix.endswith("_"):
        if index := common_prefix.rfind("_"):
            common_prefix = common_prefix[: index + 1]
        else:
            common_prefix = ""

    for literal in enum.literals:
        literal.name = literal.name.removeprefix(common_prefix)

    return common_prefix


def _process_comment(field: FieldDef) -> None:
    """Process comment of a field."""
    if match := VALID_REF_COMMENT_PATTERN.match(field.description):
        ref_msg_name, ref_const_name = match.groups()
        field.type.package = ref_msg_name
        if ref_const_name:
            field.type.name = _get_enum_identifier(
                ref_const_name.rstrip("_XXX")
            )
        else:
            field.type.name = ref_msg_name


def _get_enum_identifier(common_prefix: str) -> str:
    """Get the identifier of an enum."""
    return "".join([x.capitalize() for x in common_prefix.split("_")])


@dataclass
class MessagePkgDef:
    """Definition of a ROS message package."""

    name: str
    messages: list[MessageDef]
    packages: list[MessagePkgDef]

    def __eq__(self, other: object) -> bool:
        """Return whether the message package is equal to another."""
        if not isinstance(other, MessagePkgDef):
            return NotImplemented
        return (
            other.name == self.name
            and all(message in self.messages for message in other.messages)
            and all(package in self.packages for package in other.packages)
        )

    @classmethod
    def from_msg_folder(
        cls,
        pkg_name: str,
        msg_path: abc.AbstractFilePath | pathlib.Path,
        license_header: str | None = None,
        msg_description_regex: re.Pattern[str] | None = None,
    ) -> MessagePkgDef:
        """Create a message package definition from a folder."""
        out = cls(pkg_name, [], [])
        files = t.cast(
            t.Iterable[abc.AbstractFilePath | pathlib.Path],
            msg_path.rglob("*.msg"),
        )
        for msg_file in sorted(files, key=os.fspath):
            msg_def = MessageDef.from_file(
                msg_file, license_header, msg_description_regex
            )
            out.messages.append(msg_def)
        return out



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""The capella_ros_tools package."""
import logging
from importlib import metadata

try:
    __version__ = metadata.version("capella_ros_tools")
except metadata.PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0+unknown"
del metadata

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Tool for exporting a Capella data package to ROS messages."""

import pathlib
import re

from capellambse.metamodel import information

from capella_ros_tools import data_model

from . import logger


def _clean_name(name: str) -> str:
    return re.sub(r"\W", "", name)


def export(current_pkg: information.DataPkg, current_path: pathlib.Path):
    """Export a Capella data package to ROS messages."""
    current_path.mkdir(parents=True, exist_ok=True)
    for cls_obj in current_pkg.classes:
        fields = []
        for prop_obj in cls_obj.owned_properties:
            try:
                card = data_model.Range(
                    prop_obj.min_card.value, prop_obj.max_card.value
                )
            except AttributeError:
                card = data_model.Range("1", "1")
            type_def = data_model.TypeDef(name=prop_obj.type.name, card=card)
            prop_def = data_model.FieldDef(
                type=type_def,
                name=prop_obj.name,
                description=prop_obj.description or "",
            )
            fields.append(prop_def)
        cls_def = data_model.MessageDef(
            name=cls_obj.name,
            fields=fields,
            enums=[],
            description=cls_obj.description or "",
        )
        (current_path / f"{_clean_name(cls_obj.name)}.msg").write_text(
            str(cls_def)
        )

    for enum_obj in current_pkg.enumerations:
        literals = []
        for i, lit_obj in enumerate(enum_obj.owned_literals):
            try:
                type_name = lit_obj.value.type.name
            except AttributeError:
                type_name = "uint8"
            try:
                literal_value = lit_obj.value.value
            except AttributeError:
                literal_value = i
            type_def = data_model.TypeDef(
                type_name, data_model.Range("1", "1")
            )
            lit_def = data_model.ConstantDef(
                type=type_def,
                name=lit_obj.name,
                value=literal_value,
                description=lit_obj.description or "",
            )
            literals.append(lit_def)
        enum_def = data_model.EnumDef(
            name=enum_obj.name,
            literals=literals,
            description=enum_obj.description or "",
        )
        (current_path / f"{_clean_name(enum_obj.name)}.msg").write_text(
            str(enum_def)
        )

    for pkg_obj in current_pkg.packages:
        pkg_path = current_path / _clean_name(pkg_obj.name)
        export(pkg_obj, pkg_path)
        logger.info("Exported package %s to %s", pkg_obj.name, pkg_path)
